from pyopentsdb import utils


def aggregators(host, r_session, **kwargs):
    """
    Return all available tsdb aggregators
    :param r_session: requests.Session
    :param host: str
    :return: list
    """
    url = api_url(host)
    return utils.request_get(url, r_session, **kwargs)


def api_url(host):
    """
    Make api url for aggregators
    :param host: str
    :return: str
    """
    return '{}/api/aggregators/'.format(host)



from pyopentsdb import utils
from pyopentsdb.conf import StatsPointer


def stats(host, r_session, **kwargs):
    """
    This endpoint provides a list of statistics for the running TSD.
    Sub endpoints return details about other TSD components such as the JVM,
    thread states or storage client. All statistics are read only.

    :param host: str
    :param r_session: requests.Session
    :return: dict
    """
    return utils.request_get(api_url(host, pointer=StatsPointer.STATS), r_session, **kwargs)


def jvm(host, r_session, **kwargs):
    """
    The threads endpoint is used for debugging the TSD's JVM process and includes
    stats about the garbage collector, system load and memory usage. (v2.2)

    :param host: str
    :param r_session: requests.Session
    :return: dict
    """
    return utils.request_get(api_url(host, pointer=StatsPointer.JVM), r_session, **kwargs)


def query(host, r_session, **kwargs):
    """
    This endpoint can be used for tracking and troubleshooting queries executed
    against a TSD. It maintains an unbounded list of currently executing
    queries as well as a list of up to 256 completed queries (rotating the oldest
    queries out of memory). Information about each query includes the
    original query, request headers, response code, timing and an exception
    if thrown. (v2.2)

    :param host: str
    :param r_session: requests.Session
    :return: json
    """
    return utils.request_get(api_url(host, pointer=StatsPointer.QUERY), r_session, **kwargs)


def region_clients(host, r_session, **kwargs):
    """
    Returns information about the various HBase region server clients in AsyncHBase.
    This helps to identify issues with a particular region server. (v2.2)

    :param host: str
    :param r_session: requests.Session
    :return: json
    """
    return utils.request_get(api_url(host, pointer=StatsPointer.REGION_CLIENTS), r_session, **kwargs)


def threads(host, r_session, **kwargs):
    """
    The threads endpoint is used for debugging the TSD and providing insight
    into the state and execution of various threads without having to resort
    to a JStack trace. (v2.2)

    :param host: str
    :param r_session: requests.Session
    :return: json
    """
    return utils.request_get(api_url(host, pointer=StatsPointer.THREADS), r_session, **kwargs)


def api_url(host, pointer):
    """
    Make api url to obtain configuration

    :param host: str
    :param pointer: str
    :return: str
    """
    if pointer == StatsPointer.STATS:
        return '{}/api/stats/'.format(host)
    elif pointer == StatsPointer.JVM:
        return '{}/api/stats/jvm/'.format(host)
    elif pointer == StatsPointer.QUERY:
        return '{}/api/stats/query/'.format(host)
    elif pointer == StatsPointer.REGION_CLIENTS:
        return '{}/api/stats/region_clients/'.format(host)
    elif pointer == StatsPointer.THREADS:
        return '{}/api/stats/threads/'.format(host)



from enum import Enum


class StatsPointer(Enum):
    STATS = 1
    JVM = 2
    QUERY = 3
    REGION_CLIENTS = 4
    THREADS = 5


class ConfigPointer(Enum):
    CONFIG = 1
    FILTERS = 2


class QueryPointer(Enum):
    QUERY = 1


BASIC_TSDB_QUERY = {
        'start': None,
        'end': None,
        'msResolution': False,
        'showTSUIDs': False,
        'noAnnotations': False,
        'globalAnnotations': False,
        'showSummary': False,
        'showStats': False,
        'showQuery': False,
        'delete': False,
        'timezone': 'UTC',
        'useCalendar': False,
        'queries': list(),
    }






# Source: https://github.com/Changaco/version.py

from os.path import dirname, isdir, join
from os import environ
import re
from subprocess import CalledProcessError, check_output


PREFIX = 'v'

tag_re = re.compile(r'\btag: %s([0-9][^,]*)\b' % PREFIX)
version_re = re.compile('^Version: (.+)$', re.M)


def get_version():
    return "0.2.1"
    # Return the version if it has been injected into the file by git-archive
    version = tag_re.search('$Format:%D$')
    if version:
        return version.group(1)
    
    branch = None

    d = dirname(__file__)

    if isdir(join(d, '.git')):
        # Get the version using "git describe".
        cmd_ver = 'git describe --tags --match %s[0-9]* --dirty' % PREFIX
        cmd_branch = 'git rev-parse --abbrev-ref HEAD'
        try:
            version = check_output(cmd_ver.split()).decode().strip()[len(PREFIX):]
            branch = check_output(cmd_branch.split()).decode().strip()
        except CalledProcessError:
            raise RuntimeError('Unable to get version number from git tags')

        # PEP 440 compatibility
        # if '-' in version:
        #     if version.endswith('-dirty') and not 'IGNORE_DIRTY' in environ:
        #         raise RuntimeError('The working tree is dirty')
#            version = '.post'.join(version.split('-')[:2])

        if branch and branch != "master":
            version = "{0}.{1}".format(version, branch)
    else:
        # Extract the version from the PKG-INFO file.
        with open(join(d, 'PKG-INFO')) as f:
            version = version_re.search(f.read()).group(1)

    return version


if __name__ == '__main__':
    print(get_version())



import re

from pyopentsdb import errors
from pyopentsdb import utils


def suggest(host, r_session, **kwargs):
    try:
        t = kwargs.pop('type')
    except KeyError:
        raise errors.MissingArgumentError("'type' is a required argument")
    kwargs.update(
        {
            "data":
                {
                    'type': t,
                    'q': kwargs.pop('q', ''),
                    'max': int(kwargs.pop('max', 25))
                }
            }
        )
    return utils.request_post(api_url(host), r_session, **kwargs)


def metrics(host, r_session, **kwargs):
    kwargs.pop("type", None)
    regexp = kwargs.pop('regexp', None)

    kwargs.update({
        'type': 'metrics',
        'q': kwargs.pop('q', ''),
        'max': kwargs.pop('max', 10000),
    })

    metric_list = suggest(host, r_session, **kwargs)
    if regexp:
        metric_list = [metric for metric in metric_list if re.search(re.compile(regexp), str(metric), flags=0)]
    return metric_list


def api_url(host):
    """
    Make api url to obtain configuration
    :param host: str
    :return: str
    """
    return '{}/api/suggest/'.format(host)



from pyopentsdb import utils
from pyopentsdb.conf import ConfigPointer


def tsdb_configuration(host, r_session, **kwargs):
    return utils.request_get(api_url(host, pointer=ConfigPointer.CONFIG), r_session, **kwargs)


def filters(host, r_session, **kwargs):
    """
    :param host: str
    :param r_session: requests.Session
    :return: dict
    """
    return utils.request_get(api_url(host, pointer=ConfigPointer.FILTERS), r_session, **kwargs)


def api_url(host, pointer):
    """
    Make api url to obtain configuration

    :param host: str
    :param pointer: str
    :return: str
    """
    if pointer == ConfigPointer.CONFIG:
        return '{}/api/config/'.format(host)
    elif pointer == ConfigPointer.FILTERS:
        return '{}/api/config/filters/'.format(host)



import requests

from pyopentsdb import query
from pyopentsdb import suggest
from pyopentsdb import aggregators
from pyopentsdb import dropcaches
from pyopentsdb import config
from pyopentsdb import stats
from pyopentsdb import version
from pyopentsdb import serializers
from pyopentsdb import put


class TsdbConnector(object):
    def __init__(self, host, **kwargs):
        """
        :param port: int (mandatory); OpenTSDB instance port
        :param kwargs: see bellow
        :**kwargs options**:
        """
        self._host = str(host).rstrip('/')
        self._session = kwargs.pop("session", requests.Session())
        self._config = self._parameters_serializer()

    def _parameters_serializer(self):
        return {
            'host': self._host,
            'session': self._session
        }

    def suggest(self, **kwargs):
        """
        This endpoint provides a means of implementing an "auto-complete" call that can be accessed
        repeatedly as a user types a request in a GUI. It does not offer full text searching or wildcards,
        rather it simply matches the entire string passed in the query on the first characters
        of the stored data. For example, passing a query of type=metrics&q=sys will return the top 25
        metrics in the system that start with sys. Matching is case sensitive, so sys will not match System.CPU.
        Results are sorted alphabetically.

        :param kwargs: see bellow
        :**kwargs options**:
               * **type** * -- str;
                    The type of data to auto complete on. Must be one of the following: metrics, tagk or tagv
               * **q** * -- str;
                    A string to match on for the given type
               * **max** * -- int;
                    The maximum number of suggested results to return. Must be greater than 0
               * **<requests.request kwargs>** *
        :return: dict
        """
        return suggest.suggest(self._host, self._session, **kwargs)

    def metrics(self, **kwargs):
        """
        Helper function that give a list of available metrics
        :param kwargs: see bellow
        :**kwargs options**:
               * **max** * -- int;
                    The maximum number of suggested results to return. Must be greater than 0
               * **regxp** * -- str;
                    Regex pattern to matrics have to satisfied
               * **q** * -- str;
                    A metric to match on
               * **<requests.request kwargs>** *
        :return: dict
        """
        return suggest.metrics(self._host, self._session, **kwargs)

    def config(self, **kwargs):
        """
        This endpoint returns information about the running configuration of the TSD.
        It is read only and cannot be used to set configuration options.
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return: dict
        """
        return config.tsdb_configuration(self._host, self._session, **kwargs)

    def filters(self, **kwargs):
        """
        This endpoint lists the various filters loaded by the TSD and some information about how to use them.
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return: dict
        """
        return config.filters(self._host, self._session, **kwargs)

    def query(self, **kwargs):
        """
        Enables extracting data from the storage system in various formats determined by
        the serializer selected. Queries can be submitted via the 1.0 query string format
        or body content.

        :param kwargs: see bellow
        :**kwargs options**:
               * **start** * -- datetime;
                        The start time for the query. This can be a relative or absolute timestamp.
                        See Querying or Reading Data for details.
               * **end** * -- datetime;
                        An end time for the query. If not supplied, the TSD will assume the local system time on the server.
                        This may be a relative or absolute timestamp. See Querying or Reading Data for details.
               * **metrics** * -- dict;
                        The full name of a metric is supplied along with an optional list of tags.
                        This is optimized for aggregating multiple time series into one result.
                        aggregator: str
                                The name of an aggregation function to use. See /api/aggregators
                        metric: str
                                The name of a metric stored in the system
                        rate: bool
                                Whether or not the data should be converted into deltas before returning.
                                This is useful if the metric is a continuously incrementing counter and you
                                want to view the rate of change between data points.
                       rateOptions: dict
                                counter: bool
                                        Whether or not the underlying data is a monotonically increasing
                                        counter that may roll over
                                counterMax: int
                                        A positive integer representing the maximum value for the counter.
                                resetValue: int
                                        An optional value that, when exceeded, will cause the aggregator
                                        to return a 0 instead of the calculated rate. Useful when data sources
                                        are frequently reset to avoid spurious spikes.
                                dropResets: bool
                                        Whether or not to simply drop rolled-over or reset data points.
                       downsample: str
                                An optional downsampling function to reduce the amount of data returned.
                                examples:
                                    1h-sum
                                    30m-avg-nan
                                    24h-max-zero
                                    1dc-sum
                                    0all-sum
                                pattern:
                                    <Size><Units>-<Aggregator>-<Fill Policy>
                       tags: dict
                                To drill down to specific timeseries or group results by tag, supply one or more map values
                                in the same format as the query string. Tags are converted to filters in 2.2.
                                See the notes below about conversions. Note that if no tags are specified, all metrics
                                in the system will be aggregated into the results. Deprecated in 2.2
                                example:
                                    {tagk: tagv}
                       filters: dict
                                type: str
                                    The name of the filter to invoke. See /api/config/filters
                                tagk: str
                                    The tag key to invoke the filter on
                                filter:
                                    The filter expression to evaluate and depends on the filter being used
                                groupBy: bool
                                    Whether or not to group the results by each value matched by the filter.
                                    By default all values matching the filter will be aggregated into a single series.
                                example:
                                    filters = [
                                    {
                                        "type": "wildcard",
                                        "tagk": "host",
                                        "filter": "web01*",
                                        "groupBy": False
                                    }]
                       explicitTags: bool
                                Returns the series that include only the tag keys provided in the filters.
               * **ms** * -- bool;
                        Whether or not to output data point timestamps in milliseconds or seconds.
                        The msResolution flag is recommended. If this flag is not provided and there
                        are multiple data points within a second, those data points will be down sampled
                        using the query's aggregation function.
               * **show_tsuids** * -- bool;
                        Whether or not to output the TSUIDs associated with timeseries in the results.
                        If multiple time series were aggregated into one set, multiple TSUIDs will be
                        returned in a sorted manner
               * **no_annotations** * -- bool;
                        Whether or not to return annotations with a query. The default is to return annotations
                        for the requested timespan but this flag can disable the return. This affects both local
                        and global notes and overrides globalAnnotations
               * **global_annotations** * -- bool;
                        Whether or not the query should retrieve global annotations for the requested timespan
               * **show_summary** * -- bool;
                        Whether or not to show a summary of timings surrounding the query in the results. This creates
                        another object in the map that is unlike the data point objects. See Query Details and Stats
               * **show_stats** * -- bool;
                        Whether or not to show detailed timings surrounding the query in the results. This creates
                        another object in the map that is unlike the data point objects. See Query Details and Stats
               * **show_query** * -- bool;
                        Whether or not to return the original sub query with the query results. If the request
                        contains many sub queries then this is a good way to determine which results belong to which
                        sub query. Note that in the case of a * or wildcard query, this can produce
                        a lot of duplicate output.
               * **delete** * -- bool;
                        Can be passed to the JSON with a POST to delete any data points that match the given query.
               * **timezone** * -- str;
                        An optional timezone for calendar-based downsampling. Must be a valid timezone database name
                        supported by the JRE installed on the TSD server.
               * **use_calendar** * -- bool;
                        Whether or not use the calendar based on the given timezone for downsampling intervals
               * **<requests.request kwargs>** *
        :return: dict
        """

        return query.query(self._host, self._session, **kwargs)

    def multiquery(self, query_chunks, **kwargs):
        """
        Enables issuing multiple queries in parallel

        :param query_chunks: list; list of json serializable dicts representing OpenTSDB query
        :param kwargs: see bellow
        :**kwargs options**:
            * **max_tsdb_concurrency** -- int (optional), default=40;
                        Maximum number of concurrency threads hitting OpenTSDB api
            * **<requests.request kwargs>** *
        :return: dict
        """
        return query.multiquery(self._host, self._session,  query_chunks, **kwargs)

    def aggregators(self, **kwargs):
        """
        This endpoint simply lists the names of implemented aggregation functions used in timeseries queries.
        :return:  dict
        """
        return aggregators.aggregators(self._host, self._session, **kwargs)

    def dropcaches(self, **kwargs):
        """
        This endpoint purges the in-memory data cached in OpenTSDB. This includes all UID to name
        and name to UID maps for metrics, tag names and tag values.
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return: dict
        """
        return dropcaches.dropcaches(self._host, self._session, **kwargs)

    def serializers(self, **kwargs):
        """
        This endpoint lists the serializer plugins loaded by the running TSD. Information
        given includes the name, implemented methods, content types and methods
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return: dict
        """
        return serializers.serializers(self._host, self._session, **kwargs)

    def stats(self, **kwargs):
        """
        This endpoint provides a list of statistics for the running TSD.
        Sub endpoints return details about other TSD components such as the JVM,
        thread states or storage client. All statistics are read only.
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return: dict
        """
        return stats.stats(self._host, self._session, **kwargs)

    def jvm_stats(self, **kwargs):
        """
        The threads endpoint is used for debugging the TSD's JVM process and includes
        stats about the garbage collector, system load and memory usage. (v2.2)
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return: dict
        """
        return stats.jvm(self._host, self._session, **kwargs)

    def query_stats(self, **kwargs):
        """
        This endpoint can be used for tracking and troubleshooting queries executed
        against a TSD. It maintains an unbounded list of currently executing
        queries as well as a list of up to 256 completed queries (rotating the oldest
        queries out of memory). Information about each query includes the
        original query, request headers, response code, timing and an exception
        if thrown. (v2.2)
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return: dict
        """
        return stats.query(self._host, self._session, **kwargs)

    def region_clients(self, **kwargs):
        """
        Returns information about the various HBase region server clients in AsyncHBase.
        This helps t identify issues with a particular region server. (v2.2)
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return dict
        """
        return stats.region_clients(self._host, self._session, **kwargs)

    def threads(self, **kwargs):
        """
        The threads endpoint is used for debugging the TSD and providing insight
        into the state and execution of various threads without having to resort
        to a JStack trace. (v2.2)
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return: dict
        """
        return stats.threads(self._host, self._session, **kwargs)

    def version(self, **kwargs):
        """
        This endpoint returns information about the running version of OpenTSDB.
        :**kwargs options**:
            * **<requests.request kwargs>** *
        :return: dict
        """
        return version.version(self._host, self._session, **kwargs)

    def put(self, data, **kwargs):
        """
        Metric has to be created or enable tsd.core.auto_create_metrics
        :param data: list of dicts or dict
        :param kwargs: see bellow
        :**kwargs options**:
               * **summary** * --  bool;
                    Whether or not to return summary information
               * **details** * --  bool;
                    Whether or not to return detailed information
               * **sync** * --  bool;
                    Whether or not to wait for the data to be flushed to storage before returning the results.
               * **sync_timeout** * --  int;
                    A timeout, in milliseconds, to wait for the data to be flushed to storage before
                    returning with an error. When a timeout occurs, using the details flag will tell
                    how many data points failed and how many succeeded. sync must also be given for this
                    to take effect. A value of 0 means the write will not timeout.
               * **<requests.request kwargs>** *
        If both detailed and summary are present in a query string, the API will respond with detailed information.
        Standard status code is 204
        Standard status code with details is 200 (request._content example: b'{"success":2,"failed":0,"errors":[]}')
        :return: dict
        """
        return put.put(self._host, self._session, data, **kwargs)

    def close(self):
        self._session.close()

    def __del__(self):
        self.close()


def tsdb_connection(host, **kwargs):
    __TSDB_CONNECTION__ = TsdbConnector(
        host=host,
        **kwargs
    )
    return __TSDB_CONNECTION__



import time
import json
from pyopentsdb import errors
import requests


def make_response_message(code, message):
    return dict(
        code=int(code),
        message=message if isinstance(message, (list, dict)) else str(message)
    )


def try_contact_tsdb(exec_fn, max_attempts=3, timeout=2):
    __MAX_TSDB_CONTACT_ATTEMPTS__ = max_attempts
    __TIMEOUT__ = timeout
    n_run = 0
    result = None
    while n_run < __MAX_TSDB_CONTACT_ATTEMPTS__:
        try:
            result = exec_fn()
            if result is None:
                time.sleep(__TIMEOUT__)
                continue
            elif result.status_code in [424]:
                raise errors.FailedDependency
            return result
        except (requests.exceptions.ConnectionError, errors.FailedDependency) as e:
            if 'SSLError' in str(e):
                raise errors.SSLError(str(e))

            time.sleep(__TIMEOUT__)
            continue
        finally:
            n_run += 1

    if result is not None:
        return result
    raise errors.TsdbConnectionError("Cannot connect to TSDB host")


def request(requests_fn):
    try:
        response = try_contact_tsdb(exec_fn=requests_fn)
        if response.status_code in [200]:
            return json.loads(response.content.decode())
        elif response.status_code in [204]:
            return None
        elif response.status_code in [400]:
            raise errors.ArgumentError(json.loads(response.content.decode())["error"]["message"])
        elif response.status_code in [403]:
            raise errors.ForbiddenError(response.content.decode())
        elif response.status_code in [424]:
            raise errors.FailedDependency(json.loads(response.content.decode())["error"]["message"])
        raise errors.UncaughtError(json.loads(response.content.decode()) if response.content else 'Unknown error')
    except errors.ArgumentError:
        raise
    except errors.TsdbConnectionError:
        raise
    except errors.ForbiddenError:
        raise
    except Exception as e:
        raise errors.UncaughtError(str(e))


def request_post(url, r_session, **kwargs):
    data = json.dumps(kwargs.pop('data', dict()))
    return request(lambda: r_session.post(url, data=data, **kwargs))


def request_get(url, r_session, **kwargs):
    return request(lambda: r_session.get(url, **kwargs))



from pyopentsdb import utils


def dropcaches(host, r_session, **kwargs):
    """
    :param host: str
    :param r_session: requests.Session
    :return: dict
    """
    return utils.request_get(api_url(host), r_session, **kwargs)


def api_url(host):
    """
    Make api url for dropcaches

    :param host: str
    :return: str
    """
    return '{}/api/dropcaches/'.format(host)



from pyopentsdb import utils


def version(host, r_session, **kwargs):
    """
    This endpoint returns information about the running version of OpenTSDB.

    :param host: str
    :param r_session: requests.Session
    :return: json
    """
    return utils.request_get(api_url(host), r_session, **kwargs)


def api_url(host):
    """
    Make api url to obtain version info

    :param host: str
    :return: str
    # """
    return '{}/api/version/'.format(host)



from pyopentsdb.utils import request_post
from pyopentsdb import errors


def validate_put_data(data):
    if isinstance(data, dict):
        data = [data]

    for d in data:
        if not d.get('metric') or not d.get('timestamp') or not d.get('value') or not d.get('tags'):
            raise errors.MissingArgumentError("Missing argument/s in put data")


def put(host, r_session, data, **kwargs):
    """
    :param host: str
    :param r_session: requests.Session
    :param data: list of dicts or dict
    :param kwargs: see bellow
        :**kwargs options**:
               * **summary** * --  bool;
                    Whether or not to return summary information
               * **details** * --  bool;
                    Whether or not to return detailed information
               * **sync** * --  bool;
                    Whether or not to wait for the data to be flushed to storage before returning the results.
               * **sync_timeout** * --  int;
                    A timeout, in milliseconds, to wait for the data to be flushed to storage before
                    returning with an error. When a timeout occurs, using the details flag will tell
                    how many data points failed and how many succeeded. sync must also be given for this
                    to take effect. A value of 0 means the write will not timeout.
    :return: dict
    """

    validate_put_data(data)
    summary = kwargs.get('summary', False)
    details = kwargs.get('details', False)
    sync = kwargs.get('sync', False)
    sync_timeout = kwargs.get('sync_timeout', 0) if sync else False
    url = api_url(host, summary, details, sync, sync_timeout)
    kwargs.update(dict(data=data))
    return request_post(url, r_session, **kwargs)


def api_url(host, summary, details, sync, sync_timeout):
    url = '{}/api/put/'.format(host)
    url, previous = (''.join([url, '?summary']), True) if summary else (url, False)

    if details:
        url, previous = (''.join([url, '?details']), True) if not previous else (''.join([url, '&details']), True)
    if sync:
        url, previous = (''.join([url, '?sync']), True) if not previous else (''.join([url, '&sync']), True)
    if isinstance(sync_timeout, int) and not isinstance(sync_timeout, bool):
        url = ''.join([url, '?sync_timeout={}'.format(sync_timeout)]) \
            if not previous \
            else ''.join([url, '&sync_timeout={}'.format(sync_timeout)])
    return url



class TsdbError(Exception):
    """ TsdbError - raise on any another error related to TSDB """
    pass


class TsdbConnectionError(TsdbError):
    """ TsdbConnectionError - raise on connection error """
    pass


class TsdbQueryError(TsdbError):
    """ TsdbQueryError - raise on any query error """
    pass


class ArgumentError(TsdbError):
    """ ArgumentsError """
    pass


class MissingArgumentError(ArgumentError):
    """ MissingArgumentsError - raise when argument in json schema is missing """
    pass


class UncaughtError(TsdbError):
    """ UncaughError - raise as last possible/unknown error """
    pass


class FailedDependency(TsdbError):
    """
    FailedDependency - raise when method could not be performed on the resource because the
    requested action depended on another action and that action failed (e.g. full query)
    """
    pass


class SSLError(TsdbError):
    """
    SSLError - raise when problem with SSL
    """
    pass


class ForbiddenError(TsdbError):
    """
    ForbiddenError - raise when server return 403
    """
    pass



from queue import Queue
from queue import Empty
from threading import Thread

from pyopentsdb import errors
from pyopentsdb.utils import request_post
from pyopentsdb.conf import QueryPointer


class IterableQueue(object):
    """ Transform standard python Queue instance to iterable one"""

    def __init__(self, source_queue):
        """

        :param source_queue: queue.Queue, (mandatory)
        """
        self.source_queue = source_queue

    def __iter__(self):
        while True:
            try:
                yield self.source_queue.get_nowait()
            except Empty:
                return
            

def tsdb_query_metrics_validation(**kwargs):
    """
    looking for metric and all related and required arguments in kwargs specified in OpenTSDB http api

    :param kwargs: dict
    :return:
    """

    # tsdb query kwargs have to contain 'metrics' argument
    if not kwargs.get('metrics'):
        raise errors.MissingArgumentError("Missing argument 'metrics' in query")

    # metrics can contain more than one metric in list
    for metric_object in kwargs['metrics']:
        # each metric in metrics has to specify aggregator function
        if not metric_object.get('metric') or not metric_object.get('aggregator'):
            raise errors.MissingArgumentError("Missing argument 'metric' or 'aggregator' in metrics object")

        # each metric can contain filters
        if metric_object.get('filters'):
            for metric_filter in metric_object['filters']:
                # if filter is presented , it has contain 'type', 'tagk' and 'filter' (filter definition)
                if not metric_filter.get('type') or not metric_filter.get('tagk') or \
                        metric_filter.get('filter') is None:
                    raise errors.MissingArgumentError(
                        "Missing argument 'type', 'tagk' or 'filter' in filters object")


def query(host, r_session, **kwargs):
    """
    :param host: str
    :param r_session: requests.Session
    :param kwargs: dict
    :return: dict
    """

    # todo: make sure kwargs of tsdb are not colliding kwargs of requests

    try:
        start = kwargs.pop('start')
    except KeyError:
        raise errors.MissingArgumentError("'start' is a required argument")

    try:
        tsdb_query_metrics_validation(**kwargs)
    except errors.MissingArgumentError as e:
        raise errors.MissingArgumentError(str(e))

    # general driven arguments
    end = kwargs.pop('end', None)
    ms_resolution = bool(kwargs.pop('ms', False))
    show_tsuids = bool(kwargs.pop('show_tsuids', False))
    no_annotations = bool(kwargs.pop('no_annotations', False))
    global_annotations = bool(kwargs.pop('global_annotations', False))
    show_summary = bool(kwargs.pop('show_summary', False))
    show_stats = bool(kwargs.pop('show_stats', False))
    show_query = bool(kwargs.pop('show_query', False))
    delete_match = bool(kwargs.pop('delete', False))
    timezone = kwargs.pop('timezone', 'UTC')
    use_calendar = bool(kwargs.pop('use_calendar', False))

    queries = kwargs.pop('metrics')

    params = {
        'start': '{}'.format(int(start.timestamp())),
        'msResolution': ms_resolution,
        'showTSUIDs': show_tsuids,
        'noAnnotations': no_annotations,
        'globalAnnotations': global_annotations,
        'showSummary': show_summary,
        'showStats': show_stats,
        'showQuery': show_query,
        'delete': delete_match,
        'timezone': timezone,
        'useCalendar': use_calendar,
        'queries': list(),
    }
    if end:
        params.update({'end': int(end.timestamp())})
    params.update({'queries': queries})
    kwargs.update(dict(data=params))
    return request_post(api_url(host, pointer=QueryPointer.QUERY), r_session, **kwargs)


def multiquery(host, r_session, query_chunks, max_tsdb_concurrency=40, **kwargs):
    """
    OpenTSDB /api/query/ concurrency wrapper

    :param host: str (mandatory); OpenTSDB host
    :param r_session: requests.Session
    :param query_chunks: list (mandatory); list of json serializable dicts representing OpenTSDB query
    :param max_tsdb_concurrency: int (optional), default=40; maximum number of concurrency
                                                            threads hitting OpenTSDB api
    :return: dict; json serializable
    """

    __WORKER_RUN__ = True

    # todo: optimize, in case one of worker fail, terminate execution
    def tsdb_worker():
        while __WORKER_RUN__:
            query_kwargs = query_queue.get()

            if query_kwargs == "TERMINATOR":
                break

            # if tehre is already at least one (just one) error in queue, terminate all running threads
            # it is uselles and time consuming to finished rest of queries, if one of them fail
            if not error_queue.empty():
                break

            try:
                result = query(host, r_session, **dict(**query_kwargs, **kwargs))
                result_queue.put(result)
            except Exception as we:
                error_queue.put(we)
                break

    n_threads = min(len(query_chunks), max_tsdb_concurrency)
    query_queue = Queue(maxsize=len(query_chunks) + n_threads)
    result_queue = Queue(maxsize=len(query_chunks) + n_threads)
    error_queue = Queue()

    threads = list()
    try:
        for q in query_chunks:
            # valiate all queries in query_chunks
            tsdb_query_metrics_validation(**q)
            # add query kwargs to queue for future execution in threads
            query_queue.put(q)

        for _ in range(n_threads):
            query_queue.put("TERMINATOR")

        for _ in range(n_threads):
            t = Thread(target=tsdb_worker)
            threads.append(t)
            t.daemon = True
            t.start()

        for t in threads:
            t.join()

    except KeyboardInterrupt:
        raise
    finally:
        __WORKER_RUN__ = False

    if not error_queue.empty():
        # if not empty, error_queue has to contain exception from tsdb_worker
        raise error_queue.get()

    if result_queue.qsize() != len(query_chunks):
        # this statement is probably not necessary
        raise errors.TsdbError("Number of queries and responses is not the same")

    # make sure any other kind of response code won't be propagated to this place and will be catched and processed
    # in previous part of code
    return sum([val for val in IterableQueue(result_queue)], list())


def api_url(host, pointer):
    if pointer == QueryPointer.QUERY:
        return '{}/api/query/'.format(host)
